﻿using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Net.WebSockets;
using System.Threading;
using System.Web;


/// <summary>
/// Глобальные объекты веб приложения
/// </summary>
public class Global : HttpApplication
{
    /// <summary>
    /// Список игроков
    /// </summary>
    public static ConcurrentDictionary<string, Player> Players = new ConcurrentDictionary<string, Player>();

    /// <summary>
    /// Список комнат
    /// </summary>
    public static ConcurrentDictionary<string, Room> Rooms = new ConcurrentDictionary<string, Room>();



    /// <summary>
    /// Очередь игроков
    /// </summary>
    public static class WaitingPlayers
    {
        /// <summary>
        /// Представляет блокировщик для доступа к коллекции ожидающих игроков
        /// </summary>
        private static object locker = new object();

        /// <summary>
        /// Игроки не имеющие оппонентов
        /// </summary>
        public static List<string> queue = new List<string>();

        /// <summary>
        /// Игроки которые создали дружеское лобби
        /// </summary>
        public static List<string> friendlyQueue = new List<string>();

        /// <summary>
        /// Пытается добавить игрока в очередь.
        /// </summary>
        /// <param name="id">ID игрока</param>
        public static bool TryAdd(string id, bool friendly)
        {
            lock (locker)
            {
                try
                {
                    if (friendly)
                    {
                        // дружеское лобби
                        if (friendlyQueue.Contains(id))
                        {
                            return false;
                        }
                        else
                        {
                            friendlyQueue.Add(id);
                            return true;
                        }
                    }
                    else
                    {
                        if (queue.Contains(id))
                        {
                            return false;
                        }
                        else
                        {
                            queue.Add(id);
                            return true;
                        }
                    }
                }
                catch
                {
                    return false;
                }
            }
        }

        /// <summary>
        /// Удаляет игрока из очереди если он существует.
        /// </summary>
        /// <param name="id">ID игрока</param>
        public static bool TryRemove(string id, bool friendly)
        {
            lock (locker)
            {
                try
                {

                    if (friendly)
                    {
                        // дружеское лобби
                        if (friendlyQueue.Contains(id))
                        {
                            friendlyQueue.Remove(id);
                            return true;
                        }
                        else
                        {
                            return false;
                        }
                    }
                    else
                    {
                        if (queue.Contains(id))
                        {
                            queue.Remove(id);
                            return true;
                        }
                        else
                        {
                            return false;
                        }
                    }
                }
                catch
                {
                    return false;
                }
            }
        }

        /// <summary>
        /// Количество игроков в очереди.
        /// </summary>
        public static int Count
        {
            get
            {
                lock (locker)
                {
                    return queue.Count + friendlyQueue.Count;
                }
            }
        }



        /// <summary>
        /// Пытается подсоединиться к дружескому лобби.
        /// Обеспечивает гарантию, что игрок-инициализатор существует на момент запроса.
        /// </summary>
        /// <param name="lobbyID">ID лобби (игрока-инициализатора)</param>
        /// <returns>Возвращает успех операции</returns>
        public static bool TryGetFriendlyLobby(string lobbyID)
        {
            Monitor.Enter(locker);
            try
            {
                if (friendlyQueue.Contains(lobbyID))
                {
                    friendlyQueue.Remove(lobbyID);
                    if (Players[lobbyID].Socket.State == WebSocketState.Open)
                    {
                        return true;
                    }
                    else
                    {
                        return false;
                    }
                }
                else
                {
                    return false;
                }
            }
            catch
            {
                return false;
            }
            finally
            {
                if (Monitor.IsEntered(locker))
                    Monitor.Exit(locker);
            }
        }


        /// <summary>
        /// Возвращает первого игрока из очереди и удаляет его оттуда.
        /// Гарантирует, что сокет игрока активен на момент выдачи его из очереди.
        /// Если игрок не найден - возвращает null.
        /// </summary>
        /// <param name="who">Кто запрашивает</param>
        public static string TryGetPlayer(string who)
        {
            Monitor.Enter(locker);
            try
            {
                if (queue.Count > 0)
                {
                    int id = 0;
                    string key = queue[id];

                    // Если требуется, осуществим проверку чтобы не соединиться с ignore.
                    if (who != null)
                    {
                        if (key == who)
                        {
                            // Если самый первый равен искомому, то выдадим второго (если он существует).
                            if (queue.Count > 1)
                            {
                                id = 1;
                                key = queue[id];
                            }
                            else
                            {
                                return null;
                            }
                        }
                    }
                    // Если все корректно, то начнем процесс выдачи игрока.
                    // Удалим оппонента из очереди.
                    queue.RemoveAt(id);

                    // Если сокет оппонента впорядке.
                    if (Players[key].Socket.State == WebSocketState.Open)
                    {
                        // Удалим искавшего из очереди (если он магически не удален).
                        if (queue.Contains(who))
                        {
                            queue.Remove(who);
                        }
                        return key;
                    }
                    else
                    {
                        Monitor.Exit(locker);
                        return TryGetPlayer(who);
                    }
                }
                else
                {
                    return null;
                }
            }
            catch
            {
                return null;
            }
            finally
            {
                if (Monitor.IsEntered(locker))
                    Monitor.Exit(locker);
            }
        }
    }

}