﻿
using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System.Threading;

/// <summary>
/// Комната для игры.
/// </summary>
public class Room
{
    /// <summary>
    /// Рандомизатор
    /// </summary>
    static Random Rand = new Random();

    /// <summary>
    /// ID комнаты.
    /// </summary>
    public string ID { get; set; }

    /// <summary>
    /// Первый игрок.
    /// </summary>
    public string FirstPlayer { get; set; }

    /// <summary>
    /// Второй игрок.
    /// </summary>
    public string SecondPlayer { get; set; }



    /// <summary>
    /// Последний ход первого игрока.
    /// </summary>
    public int[] LastStepByFirst { get; set; }
    /// <summary>
    /// Последний ход второго игрока.
    /// </summary>
    public int[] LastStepBySecond { get; set; }



    /// <summary>
    /// Показывает совершен ли первый ход.
    /// Необходимо для генерации минного поля после первого хода.
    /// </summary>
    public bool firstStep = true;

    /// <summary>
    /// Показывает, на какой мине взорвался игрок
    /// </summary>
    public int[] explodedAt = { 0, 0 };

    /// <summary>
    /// Размер поля по X.
    /// </summary>
    const int sizeX = 30;
    /// <summary>
    /// Размер поля по Y.
    /// </summary>
    const int sizeY = 20;
    /// <summary>
    /// Процент минных клеток. Оптимально ~ 24.
    /// </summary>
    const double minesProcent = 24D;

    /// <summary>
    /// Минное поле
    /// </summary>
    public int[,] Field { get; set; }

    /// <summary>
    /// Чей ход в данный момент
    /// </summary>
    public string TurnBy { get; set; }

    /// <summary>
    /// Таймаут для хода
    /// </summary>
    const int TurnTimeout = 25;

    /// <summary>
    /// Таймер ходов игроков
    /// </summary>
    System.Timers.Timer TurnTimer = new System.Timers.Timer(TurnTimeout * 1000);

    /// <summary>
    /// Количество мин
    /// </summary>
    int MinesCount
    {
        get
        {
            return (int)Math.Floor((sizeX * sizeY) * (minesProcent / 100) + 0.5);
        }
    }

    /// <summary>
    /// Открытые клетки минного поля.
    /// Элементы соответственно: { X; Y; Сколько_мин_вокруг }
    /// </summary>
    public List<int[]> OpenCells = new List<int[]>();

    /// <summary>
    /// Определяет, была ли уже кликнута эта клетка
    /// </summary>
    /// <param name="pos">Позиция клетки</param>
    public bool CellsIsAlreadyClicked(int x, int y)
    {
        bool found = false;
        foreach (int[] t in OpenCells)
        {
            if (t[0] == x && t[1] == y)
            {
                found = true;
                break;
            }
        }
        return found;
    }


    /// <summary>
    /// Получает время в формате Unix-Time
    /// </summary>
    /// <returns></returns>
    public int GetTimeStamp()
    {
        return (int)(DateTime.UtcNow.Subtract(new DateTime(1970, 1, 1))).TotalSeconds;
    }



    /// <summary>
    /// Какой-либо игрок покинул/проиграл игру.
    /// Сброс параметров и удаление комнаты.
    /// </summary>
    public async Task SecondPlayerWasLeave()
    {
        await Task.Run(() =>
        {
            // Сбросим комнату у второго игрока
            if (Global.Players.ContainsKey(FirstPlayer))
            {
                Global.Players[FirstPlayer].RoomDeletedEvent();
            }
            if (Global.Players.ContainsKey(SecondPlayer))
            {
                Global.Players[SecondPlayer].RoomDeletedEvent();
            }
            if (Global.Rooms.ContainsKey(ID))
            {
                Room toDelete;
                Global.Rooms.TryRemove(ID, out toDelete);
            }
        });
    }

    private object OpenCell_Locker = new object();

    /// <summary>
    /// Сообщает игрокам состояние комнаты.
    /// </summary>
    /// <param name="itEnd">Флаг "игра завершена".</param>
    public async Task SendRoomInfo(bool itEnd)
    {
        try
        {
            await Task.Run(() =>
            {
                var data = new JObject();
                data["roomID"] = ID;
                data["fieldSize"] = sizeX + "x" + sizeY;
                data["minesCount"] = MinesCount.ToString();
                lock (OpenCell_Locker)
                {
                    data["openCells"] = JToken.FromObject(OpenCells);
                }
                data["lastB"] = string.Empty;
                data["opponentNick"] = Global.Players[SecondPlayer].Nickname;
                data["timestamp"] = GetTimeStamp();
                data["itEnd"] = itEnd;
                if (itEnd)
                {
                    data["explodedAt"] = JToken.FromObject(explodedAt);
                }
                data["turnBy"] = TurnBy;

                var answer = new Сommunication.Responce(
                    Сommunication.MessageType.Status, "RoomInformation", data
                    );

                //try
                //{
                if (LastStepBySecond != null)
                    data["lastB"] = JToken.FromObject(LastStepBySecond);
                data["timeoutsS"] = Global.Players[FirstPlayer].TurnPassInSequence;
                data["timeoutsA"] = Global.Players[FirstPlayer].TurnPassAll;
                Global.Players[FirstPlayer].SendData(answer);

                if (LastStepByFirst != null)
                    data["lastB"] = JToken.FromObject(LastStepByFirst);
                data["opponentNick"] = Global.Players[FirstPlayer].Nickname;
                data["timeoutsS"] = Global.Players[SecondPlayer].TurnPassInSequence;
                data["timeoutsA"] = Global.Players[SecondPlayer].TurnPassAll;
                Global.Players[SecondPlayer].SendData(answer);
                /* }
                 catch
                 {
                     SendExceptionToPlayers(new Exception("Connection was lost."));
                 }*/
            });
        }
        catch (Exception e)
        {
            SendExceptionToPlayers(e);
        }
    }

    /// <summary>
    /// Возвращает кол-во мин вокруг клетки
    /// </summary>
    /// <param name="x">Позиция по X</param>
    /// <param name="y">Позиция по Y</param>
    public int GetMinesAround(int x, int y)
    {
        int count = 0;
        for (int i = -1; i < 2; ++i)
            for (int j = -1; j < 2; ++j)
            {
                if (i == 0 && j == 0)
                    continue;
                int posX = x + i, posY = y + j;
                if ((posX >= 0 && posX < sizeX) && (posY >= 0 && posY < sizeY))
                    if (Field[posX, posY] == 1)
                        ++count;
            }
        return count;
    }


    /// <summary>
    /// Открывает рекурсивно клетки вокруг пустой
    /// </summary>
    /// <param name="x">Позиция по X</param>
    /// <param name="y">Позиция по Y</param>
    public void OpenSubCell(int x, int y)
    {
        if (GetMinesAround(x, y) == 0)
        {
            for (int i = -1; i < 2; ++i)
                for (int j = -1; j < 2; ++j)
                {
                    if (i == 0 && j == 0)
                        continue;
                    int posX = x + i, posY = y + j;
                    if ((posX >= 0 && posX < sizeX) && (posY >= 0 && posY < sizeY))
                    {
                        if (!CellsIsAlreadyClicked(posX, posY))
                        {
                            OpenCells.Add(new int[] { posX, posY, GetMinesAround(posX, posY) });
                            if (GetMinesAround(posX, posY) == 0)
                            {
                                OpenSubCell(posX, posY);
                            }
                        }
                    }
                }
        }
    }

    /// <summary>
    /// Разрешить ход следующему игроку
    /// </summary>
    public void NextPlayerTurn()
    {
        TurnBy = TurnBy == FirstPlayer ? SecondPlayer : FirstPlayer;
    }

    /// <summary>
    /// Проверяет, полностью ли совпадают ли флажки с бомбами
    /// </summary>
    public bool IsWinFlags(int[][] cells)
    {
        try
        {
            int countTrue = 0;
            int countFalse = 0;
            foreach (int[] p in cells)
            {
                if (Field[p[0], p[1]] == 1)
                    ++countTrue;
                else
                    ++countFalse;
            }
            return countTrue == MinesCount && countFalse == 0;
        }
        catch
        {
            return false;
        }

    }

    /// <summary>
    /// Отправляет игрокам хинт
    /// </summary>
    /// <param name="x">Координата хинта по X</param>
    /// <param name="y">Координата хинта по Y</param>
    public async Task SendHint(int x, int y)
    {
        await Task.Run(() =>
        {
            if ((x >= 0 && x < sizeX) && (y >= 0 && y < sizeY))
            {
                var data = new JObject();
                data["x"] = x;
                data["y"] = y;
                var answer = new Сommunication.Responce(Сommunication.MessageType.Command, "NewHint", data);
                Global.Players[FirstPlayer].SendData(answer);
                Global.Players[SecondPlayer].SendData(answer);
            }
        });
    }

    /// <summary>
    /// Определяет, являются ли флажки полностью корректными
    /// </summary>
    /// <param name="who">Запросивший игрок</param>
    /// <param name="flags">Флажки</param>
    /// <returns></returns>
    public async Task TestFlags(Player who, int[][] flags)
    {
        await Task.Run(() =>
        {
            if (flags == null)
            {
                var data = new JObject();
                data["text"] = "The flags are not set.";
                var answer = new Сommunication.Responce(
                    Сommunication.MessageType.Error, string.Empty, data
                    );
                who.SendData(answer);
            }
            else
            {
                try
                {
                    // Если магическим образом ещё не нажали ни разу - сгенерируем поле!
                    if (firstStep) FieldGenerate(0, 0);
                    int[][] fl = flags;
                    if (IsWinFlags(fl))
                    {
                        // Победа! Все флажки корректны.
                        EndOfGame(who.ID, "All the mines have been marked.");
                    }
                    else
                    {
                        // Флажки не корректны.
                        var data = new JObject();
                        data["text"] = "List of flags isn't valid.";
                        var answer =
                        new Сommunication.Responce(Сommunication.MessageType.Error, string.Empty, data);
                        who.SendData(answer);
                    }
                }
                catch (JsonException)
                {
                    var data = new JObject();
                    data["text"] = "Invalid format flags list!";
                    var answer =
                    new Сommunication.Responce(Сommunication.MessageType.Error, string.Empty, data);
                    who.SendData(answer);
                }
                catch (Exception e)
                {
                    var data = new JObject();
                    data["text"] = e.Message;
                    var answer =
                    new Сommunication.Responce(Сommunication.MessageType.ServerError, string.Empty, data);
                    who.SendData(answer);
                }
            }

        });
    }


    /// <summary>
    /// Превращает все не заминированные клетки в открытые
    /// </summary>
    public void AllEmptyCellsToOpen()
    {
        lock (OpenCell_Locker)
        {
            OpenCells.Clear();
            for (int x = 0; x <= Field.GetUpperBound(0); ++x)
            {
                for (int y = 0; y <= Field.GetUpperBound(1); ++y)
                {
                    if (Field[x, y] == 0)
                    {
                        OpenCells.Add(new int[] { x, y, GetMinesAround(x, y) });
                    }
                }
            }
        }
    }

    public async Task EndOfGame(string winnerID, string cause)
    {
        await Task.Run(() =>
        {
            AllEmptyCellsToOpen();
            Task.WaitAll(SendRoomInfo(true));

            string dataForFirst = "You win!";
            string dataForSecond = "You win!";

            if (FirstPlayer == winnerID)
            {
                dataForSecond = "You lose!";
            }
            else
            {
                dataForFirst = "You lose!";
            }
            var data = new JObject();
            data["text"] = dataForFirst;
            data["cause"] = cause;
            var answer =
            new Сommunication.Responce(Сommunication.MessageType.Command, "PlayerStatus", data);
            Global.Players[FirstPlayer].SendData(answer);
            data["text"] = dataForSecond;
            answer =
            new Сommunication.Responce(Сommunication.MessageType.Command, "PlayerStatus", data);
            Global.Players[SecondPlayer].SendData(answer);


            // Очистим ресурсы комнаты.
            SecondPlayerWasLeave();
        });
    }


    public async Task ToChat(Player who, string msg)
    {
        await Task.Run(() =>
        {
            var data = new JObject();
            data["msg"] = msg;
            var answer =
            new Сommunication.Responce(Сommunication.MessageType.Command, "NewMsg", data);
            if (who.ID == FirstPlayer)
            {
                Global.Players[SecondPlayer].SendData(answer);
            }
            else
            {
                Global.Players[FirstPlayer].SendData(answer);
            }
        });
    }


    public async Task TimeOut()
    {
        try
        {
            ++Global.Players[TurnBy].TurnPassInSequence;
            ++Global.Players[TurnBy].TurnPassAll;
            if (Global.Players[TurnBy].TurnPassInSequence > 2)
            {
                EndOfGame(TurnBy == FirstPlayer ? SecondPlayer : FirstPlayer, "Triple timeout.");
                return;
            }
            if (Global.Players[TurnBy].TurnPassAll > 5)
            {
                EndOfGame(TurnBy == FirstPlayer ? SecondPlayer : FirstPlayer, "Sixth timeout.");
                return;
            }
            NextPlayerTurn();
            SendRoomInfo(false);
        }
        catch (Exception e)
        {
            SendExceptionToPlayers(e);
        }
    }


    /// <summary>
    /// Определяет корректность миддл-фрагов вокруг клетки
    /// </summary>
    /// <param name="cells">Клетки которые необходимо открыть</param>
    /// <param name="byX">Координата центровой клетки по X</param>
    /// <param name="byY">Координата центровой клетки по Y</param>
    /// <returns></returns>
    public bool IsOpenCellsCorrect(int[][] cells, string byX, string byY)
    {
        int _x, _y;
        if (int.TryParse(byX, out _x) && int.TryParse(byY, out _y) && _x >= 0 && _y >= 0)
        {
            if (cells.Length <= 8)
            {
                bool allOK = true;
                for (int i = 0; i < cells.Length; ++i)
                {
                    if (cells[i].Length == 2)
                    {
                        if (Math.Abs(cells[i][0] - _x) > 1 || Math.Abs(cells[i][1] - _y) > 1)
                        {
                            allOK = false;
                            break;
                        }
                    }
                    else
                    {
                        allOK = false;
                        break;
                    }
                }
                return allOK;
            }
            else
                return false;
        }
        else
        {
            return false;
        }
    }


    /// <summary>
    /// Открывает клетку
    /// </summary>
    /// <param name="who">Запросивший игрок</param>
    /// <param name="x">Позиция по X</param>
    /// <param name="y">Позиция по Y</param>
    public async Task OpenCell(Player who, int[][] cells, string byX, string byY)
    {
        await Task.Run(() =>
        {
            if (TurnBy == who.ID && IsOpenCellsCorrect(cells, byX, byY))
            //if(true)
            {
                int _x = 0;
                int _y = 0;

                // Начинаем обработку клеток
                bool allCorrect = true;
                var openNow = new List<int[]>();

                for (int i = 0; i < cells.Length; ++i)
                {
                    if (cells[i].Length == 2)
                    {
                        // Если координаты - целые числа
                        if (int.TryParse(cells[i][0].ToString(), out _x)
                        && int.TryParse(cells[i][1].ToString(), out _y))
                        {
                            // Если позиция клетки корректна и ход наш.
                            if (_x >= 0 && _x < sizeX && _y >= 0 && _y < sizeY)
                            {
                                // Если первый ход в игре
                                if (firstStep)
                                {
                                    // Генерируем минное поле
                                    FieldGenerate(_x, _y);
                                }
                                if (CellsIsAlreadyClicked(_x, _y))
                                {
                                    var data = new JObject();
                                    data["text"] = "The cell is already open!";
                                    var answer = new Сommunication.Responce(
                                        Сommunication.MessageType.Error, string.Empty, data
                                        );
                                    who.SendData(answer);
                                    allCorrect = false;
                                    break;
                                }
                                else
                                {
                                    // Проверим есть ли мина в клетке.
                                    if (Field[_x, _y] == 1)
                                    {
                                        // Если есть - разошлем обоим игрокам уведомления о конце игры.
                                        explodedAt = new int[] { _x, _y };
                                        string winnerID = FirstPlayer;
                                        if (FirstPlayer == who.ID)
                                        {
                                            winnerID = SecondPlayer;
                                        }
                                        EndOfGame(winnerID, "A mine has exploded.");
                                        allCorrect = false;
                                        break;
                                    }
                                    else
                                    {
                                        try
                                        {
                                            // Иначе откроем клетку и разрешим ход другому игроку.
                                            lock (OpenCell_Locker)
                                            {
                                                int[] t = new int[] { _x, _y, GetMinesAround(_x, _y) };
                                                OpenCells.Add(t);
                                                openNow.Add(t);
                                            }
                                        }
                                        catch (Exception e)
                                        {
                                            SendExceptionToPlayers(e);
                                            allCorrect = false;
                                            break;
                                        }
                                    }
                                }
                            }
                            else
                            {
                                var data = new JObject();
                                data["text"] = "The position of the cell is outside the field!";
                                var answer = new Сommunication.Responce(
                                    Сommunication.MessageType.Error, string.Empty, data
                                    );
                                who.SendData(answer);
                                allCorrect = false;
                                break;
                            }
                        }
                        else
                        {
                            var data = new JObject();
                            data["text"] = "The position of the cell isn't integer or not set!";
                            var answer = new Сommunication.Responce(
                                Сommunication.MessageType.Error, string.Empty, data
                                );
                            who.SendData(answer);
                            allCorrect = false;
                            break;
                        }
                    }
                    else
                    {
                        var data = new JObject();
                        data["text"] = "Incorrect cells!";
                        var answer = new Сommunication.Responce(
                            Сommunication.MessageType.Error, string.Empty, data
                            );
                        who.SendData(answer);
                        allCorrect = false;
                        break;
                    }

                }
                // Если все открытые клетки были корректны
                if (allCorrect)
                {
                    // Иначе откроем клетку и разрешим ход другому игроку.
                    TurnTimer.Stop();
                    for (int i = 0; i < openNow.Count; ++i)
                    {
                        OpenSubCell(openNow[i][0], openNow[i][1]);
                    }
                    Global.Players[TurnBy].TurnPassInSequence = 0;
                    NextPlayerTurn();
                    if (who.ID == FirstPlayer)
                    {
                        LastStepByFirst = new int[] { _x, _y };
                    }
                    else
                    {
                        LastStepBySecond = new int[] { _x, _y };
                    }
                    SendRoomInfo(false);
                    TurnTimer.Start();
                }
            }
            else
            {
                var data = new JObject();
                data["text"] = "The turn isn't yours!";
                var answer = new Сommunication.Responce(
                    Сommunication.MessageType.Error, string.Empty, data
                    );
                who.SendData(answer);
            }
        });
    }






    /// <summary>
    /// Генерирует минное поле, исключающее первую клетку
    /// </summary>
    /// <param name="nullX">Исключающее X</param>
    /// <param name="nullY">Исключающее Y</param>
    public void FieldGenerate(int nullX, int nullY)
    {
        firstStep = false;

        Field = new int[sizeX, sizeY];
        List<int> minesRandomizer = new List<int>();
        for (int i = 0; i < sizeX * sizeY; ++i)
        {
            minesRandomizer.Add(i);
        }
        for (int i = 0; i < MinesCount; ++i)
        {
            while (true)
            {
                int id = Rand.Next(0, minesRandomizer.Count);
                int randCell = minesRandomizer[id];
                int posX = (int)Math.Floor((double)randCell / sizeY);
                int posY = randCell - posX * sizeY;
                // Если не заменировано все поле
                if (MinesCount < sizeX * sizeY)
                {
                    if (posX == nullX && posY == nullY) continue;
                }
                minesRandomizer.RemoveAt(id);
                Field[posX, posY] = 1;
                break;
            }
        }

        // Тут магия чтобы в первый раз всегда открывалось как минимум 2 пустоты...
        int[] t = new int[] { nullX, nullY, GetMinesAround(nullX, nullY) };
        int c = 0;
        bool likeX = false;
        bool likeY = false;
        lock (OpenCell_Locker)
        {
            OpenCells.Add(t);
            OpenSubCell(t[0], t[1]);
            int pred = -1;
            for (int i = 0; i < OpenCells.Count; ++i)
            {
                if (OpenCells[i][2] == 0)
                {
                    ++c;
                    if (i > 0 && pred > 0)
                    {
                        // Делаем чтобы небыло просто вертикальных/горизонтальных пустот
                        if (OpenCells[i][0] != OpenCells[pred][0])
                            likeX = true;
                        if (OpenCells[i][1] != OpenCells[pred][1])
                            likeY = true;
                    }
                    pred = i;
                }

            }
            OpenCells.Clear();
        }
        if (c < 2 || c > 12 || !(likeX & likeY))
        {
            FieldGenerate(nullX, nullY);
        }

    }

    public void SendExceptionToPlayers(Exception e)
    {
        var aa = new JObject();
        aa["messageAA"] = e.Message;
        aa["messageAB"] = e.Source;
        aa["messageAC"] = e.StackTrace;
        aa["messageAD"] = e.TargetSite.Name;
        aa["f"] = FirstPlayer;
        aa["s"] = SecondPlayer;
        var bb = new Сommunication.Responce(Сommunication.MessageType.ServerError, "Expection", aa);
        Global.Players[FirstPlayer].SendData(bb);
        Global.Players[SecondPlayer].SendData(bb);
    }

    /// <summary>
    /// Основной конструтор комнаты.
    /// </summary>
    /// <param name="firstPlayer">Первый игрок</param>
    /// <param name="secondPlayer">Второй игрок</param>
    /// <param name="roomID">ID комнаты (равен ключу в словаре)</param>
    public Room(string firstPlayer, string secondPlayer, string roomID)
    {
        FirstPlayer = firstPlayer;
        SecondPlayer = secondPlayer;
        try
        {
            ID = roomID;
            TurnBy = (Rand.Next(0, 2) == 0) ? FirstPlayer : SecondPlayer;
            if (Global.Players.ContainsKey(FirstPlayer) && Global.Players.ContainsKey(SecondPlayer))
            {
                Global.Players[FirstPlayer].RoomID = ID;
                Global.Players[SecondPlayer].RoomID = ID;
                TurnTimer.Elapsed += async (sender, e) => TimeOut();
                SendRoomInfo(false);
                TurnTimer.Start();

            }
            else
            {
                SecondPlayerWasLeave();
            }
        }

        catch (Exception e)
        {
            SendExceptionToPlayers(e);
        }
    }
}